import re, sys, os
try:
    from urllib.parse import unquote, unquote_plus
except ImportError:
    from urlparse import unquote
    from urllib import unquote_plus
import json
from hashlib import md5
import random

####################################################################################
# Utitlities
####################################################################################
REQUEST = re.compile ('([^ ]+) ([^ ]+)(( HTTP/([0-9.]+))$|$)')
CONNECTION = re.compile ('Connection: (.*)', re.IGNORECASE)

def crack_query (r):
    if type (r) is bytes:
        r = r.decode ("utf8")

    if not r: return {}
    if r[0]=='?': r=r[1:]
    arg={}
    q = [x.split('=', 1) for x in r.split('&')]

    for each in q:
        k = unquote_plus (each [0])
        try:
            t, k = k.split (":", 1)
        except ValueError:
            t = "str"

        v = unquote_plus (each [1])
        if len (each) == 2:
            if t == "str":
                pass
            elif t in ("i", "int"):
                v = int (v)
            elif t in ("n", "null"):
                v = None
            elif t in ("f", "float"):
                v = float (v)
            elif t in ("l", "list"):
                v = v.split (",")
            elif t in ("b", "bool"):
                v = v.lower () in ("1", "true", "yes")

        if k in arg:
            if type (arg [k]) is not type ([]):
                arg[k] = [arg[k]]
            arg[k].append (v)

        else:
            arg[k] = v

    return arg

def crack_request (r):
    m = REQUEST.match (r)
    if m.end() == len(r):
        uri=m.group(2)
        if m.group(3):
            version = m.group(5)
        else:
            version = None
        return m.group(1).lower(), uri, version
    else:
        return None, None, None

def join_headers (headers):
    r = []
    for i in range(len(headers)):
        if headers[i][0] in ' \t':
            r[-1] = r[-1] + headers[i][1:]
        else:
            r.append (headers[i])
    return r

def get_header (head_reg, lines, group=1):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m.group (group)
    return ''

def get_header_match (head_reg, lines):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m
    return ''

def get_extension (path):
    dirsep = path.rfind ('/')
    dotsep = path.rfind ('.')
    if dotsep > dirsep:
        return path[dotsep+1:]
    else:
        return ''

ALNUM = '0123456789abcdefghijklmnopqrstuvwxyz'
def md5uniqid (length = 13):
    global ALNUM
    _id = ''
    for i in range (0, length):
        _id += random.choice(ALNUM)
    return md5 (_id.encode ("utf8")).hexdigest ()[length:]

def parse_params (header_val):
    d = {}
    header_filed = ""
    last_key = None
    for each in header_val.split (";"):
        if last_key:
            d [last_key] += ";" + each
            if each [-1] == '"':
                last_key = None
            continue

        each = each.strip ()
        if not each: continue

        try:
            a, b = each.split ("=", 1)
        except ValueError:
            if header_filed:
                header_filed += ":" + each
            else:
                header_filed = each
            continue

        key = a.lower ()
        d [key] = b
        if len (b) >= 2 and  b [0] == '"' and b [-1] != '"':
            last_key = key

    return header_filed, d

def parse_range (rg, file_length):
    if rg.startswith ('bytes='):
        s, e = rg [6:].split ("-", 1)
    else:
        return 0, file_length - 1

    s = int (s)
    if e:
        e = int (e)
    else:
        e = min (s + 1048575, file_length - 1)
    assert e < file_length
    assert s <= e
    return s, e

def parse_multi_params (header_val):
    return dict ([parse_params (v) for v in header_val.split (",")])



import mimetypes
import os

"""-----------------------------286502649418924
Content-Disposition: form-data; name="submit-hidden"

Genious
-----------------------------286502649418924
Content-Disposition: form-data; name="submit-name"

Hans Roh
-----------------------------286502649418924
Content-Disposition: form-data; name="file1"; filename="House.2x01.acceptance.DVDRip-iMa.smi"
Content-Type: application/smil

asdadasda
-----------------------------286502649418924
Content-Disposition: form-data; name="file2"; filename=""
Content-Type: application/octet-stream


-----------------------------286502649418924--"""


class multipart_producer:
	buffer_size = 4096
	
	def __init__ (self, data, boundary):
		# self.data = {"name": "Hans Roh", "file1": <open (path, "rb")>}
		self.data = data
		self.closed = False
		self.serialized = []
		self.dlist = []
		self.boundary = boundary		
		self.current_file = None
		self.content_length = self.calculate_size ()
		#self.bytes_out = 0
		
	def calculate_size (self):
		size = (len (self.boundary.encode ("utf8")) + 4) * (len (self.data) + 1) # --boundary + (\r\n or last --)
		for name, value in list(self.data.items ()):
			size += (41 + len (name.encode ("utf8"))) #Content-Disposition: form-data; name=""\r\n
			if type (value) is not type (""):
				fsize = os.path.getsize (value.name)
				size += (fsize + 2) # file size + \r\n
				fn = os.path.split (value.name) [-1]
				size += len (fn.encode ("utf8")) + 13 # ; filename=""\r\n
				mimetype = mimetypes.guess_type (fn) [0]
				if not mimetype:
					mimetype = "application/octet-stream"
				size += (16 + len (mimetype)) # Content-Type: application/octet-stream\r\n				
				self.dlist.append ((1, name, (value.name, fn, mimetype)))
				value.close ()
			
			else:
				size += len (value.encode ("utf8")) + 2 # value + \r\n
				self.dlist.append ((0, name, value))

			size += 2 # header end \r\n	
			
		return size	

	def get_content_length (self):
		return self.content_length
	
	def get_payload (self):
		if not self.serialized:
			raise ValueError ('Expired Multipart Content')
		return b"".join (self.serialized)
	
	def exhausted (self):
		return self.closed
	
	def more (self):
		d = self.__more ()
		if d and self.content_length <= 4096000: # MAX 4M
			self.serialized.append (d)
		return d
			
	def __more (self):
		if not self.dlist: 
			self.closed = True
			return b''
			
		if self.current_file:
			d = self.current_file.read (self.buffer_size)
			if d:
				return d
			else:
				self.current_file.close ()
				self.current_file = None
				self.dlist.pop (0)
				d = "\r\n"
				if not self.dlist:
					d += "--%s--" % self.boundary
				#self.bytes_out += len (d)
				return d.encode ("utf8")
			
		if self.dlist:
			first = self.dlist [0]
			if first [0] == 0: # formvalue
				d = '--%s\r\nContent-Disposition: form-data; name="%s"\r\n\r\n%s\r\n' % (self.boundary, first [1], first [2])
				self.dlist.pop (0)
				if not self.dlist:
					d += "--%s--" % self.boundary
				#self.bytes_out += len (d)
				return d.encode ("utf8")
			else:
				path, filename, mimetype = first [2]
				self.current_file = open (path, "rb")
				return ('--%s\r\nContent-Disposition: form-data; name="%s"; filename="%s"\r\nContent-Type: %s\r\n\r\n' % (
					self.boundary, first [1], filename, mimetype
				)).encode ("utf8")
				#self.bytes_out += len (d)
				#return d
		


from rs4 import asyncore
import re
import time
import sys
import threading
from ..client.asynconnect import DEFAULT_ZOMBIE_TIMEOUT, DEFAULT_KEEP_ALIVE

DEBUG = False

class OperationalError (Exception):
	pass

class SQLError (Exception):
	pass

class DBConnect:
	zombie_timeout = DEFAULT_ZOMBIE_TIMEOUT
	keep_alive = DEFAULT_KEEP_ALIVE

	def __init__ (self, address, params = None, lock = None, logger = None):
		self.address = address
		self.params = params
		self.dbname, self.user, self.password = "", "", ""

		if self.params:
			self.dbname, auth = self.params
			if auth:
				if type (auth) is not tuple:
					self.user = auth
				else:
					try: self.user, self.password = auth
					except ValueError: self.username = auth

		self.lock = lock
		self.logger = logger

		self.request_count = 0
		self.execute_count = 0

		self._cv = threading.Condition ()
		self.backend = False

		# need if there's not any request yet
		self.active = 0
		self.request = None
		self.has_result = False
		self.__history = []
		self.__no_more_request = False

		self.set_event_time ()

	def close_if_over_keep_live (self):
		if self.connected and time.time () - self.event_time > self.keep_alive:
			self.disconnect ()

	def set_backend (self, backend_keep_alive = 10):
		self.backend = True
		self.keep_alive = backend_keep_alive

	def duplicate (self):
		new_asyncon = self.__class__ (self.address, self.params, self.lock, self.logger)
		new_asyncon.keep_alive = self.keep_alive
		new_asyncon.backend = self.backend
		return new_asyncon

	def get_proto (self):
		# call by culster_manager
		return None

	def handle_abort (self):
		# call by dist_call
		self.close ()

	def close (self, deactive = 1):
		addr = type (self.address) is tuple and ("%s:%d" % self.address) or str (self.address)
		if deactive:
			self.logger ("[info] ..dbo %s has been closed" % addr)
			self.set_active (False)
			self.request = None
		#else:
		#	self.logger ("[info] ..dbo %s was disconnected or keep alive timeout" % addr)

	def get_history (self):
		return self.__history

	def clean_shutdown_control (self, phase, time_in_this_phase):
		self.__no_more_request = True
		if self.isactive ():
			return 1
		else:
			self.close ()
			self.__no_more_request = False
			return 0

	def maintern (self, object_timeout):
		# when in map, mainteren by lifetime with zombie_timeout
		if self.is_channel_in_map ():
			return False
		idle = time.time () - self.event_time
		if idle > object_timeout:
			self.close (1)
			return True # deletable
		return False

	def reconnect (self):
		self.disconnect ()
		self.connect ()

	def disconnect (self):
		# keep request, just close temporary
		self.close (deactive = 0)

	def close_case (self):
		raise NotImplementedError

	def set_active (self, flag, nolock = False):
		if not flag:
			self.set_timeout (self.keep_alive)

		if flag:
			flag = time.time ()
		else:
			flag = 0

		if nolock or self.lock is None:
			self.active = flag
			return

		with self.lock:
			self.active = flag
			self.request_count += 1
			if not flag:
				self.has_result = False

	def get_active (self, nolock = False):
		if nolock or self.lock is None:
			return self.active
		self.lock.acquire ()
		active = self.active
		self.lock.release ()
		return active

	def isactive (self):
		return self.get_active () > 0

	def isconnected (self):
		# self.connected should be defined at __init__ or asyncore
		return self.connected

	def get_request_count (self):
		return self.request_count

	def get_execute_count (self):
		return self.execute_count

	def connect (self, force = 0):
		raise NotImplementedError("must be implemented in subclass")

	def set_timeout (self, timeout):
		self.zombie_timeout = timeout

	def handle_timeout (self):
		self.handle_close (OperationalError ("Operation Timeout"))

	def handle_error (self):
		dummy, t, v, info = asyncore.compact_traceback ()
		self.has_result = False
		self.logger.trace ()
		self.handle_close (v)

	def handle_close (self, expt = None):
		if self.expt is None:
			self.expt = expt
		self.close_case_with_end_tran ()
		self.close ()

	def set_event_time (self):
		self.event_time = time.time ()

	def log_history (self, msg):
		self.__history.append ("BEGIN TRAN: %s" % sql)

	def close_case_with_end_tran (self):
		self.end_tran ()
		self.close_case ()

	#-----------------------------------------------------
	# DB methods
	#-----------------------------------------------------
	def fetchall (self):
		raise NotImplementedError

	def end_tran (self):
		raise NotImplementedError

	def begin_tran (self, request):
		if self.__no_more_request:
			raise OperationalError ("Entered Shutdown Process")
		self.request = request
		self.__history = []
		self.out_buffer = ''
		self.has_result = False
		self.expt = None
		self.execute_count += 1
		self.close_if_over_keep_live ()
		self.set_event_time ()

	def execute (self, request):
		self.begin_tran (request)
		raise NotImplementedError("must be implemented in subclass")


class AsynDBConnect (DBConnect):
	def is_channel_in_map (self, map = None):
		if map is None:
			map = self._map
		return self._fileno in map

	def del_channel (self, map=None):
		# do not remove self._fileno
		fd = self._fileno
		if map is None:
			map = self._map
		if fd in map:
			del map[fd]




import threading, random
from . import asynconnect, synconnect
import time
try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse	
import copy
from ..protocols.http2 import H2_PROTOCOLS
from ..protocols import http2

PROTO_CONCURRENT_STREAMS = H2_PROTOCOLS

def select_channel (asyncons):
	asyncon = None
	_asyncons = []
	
	for each in asyncons:
		if each.isactive ():
			continue
		if not each.isconnected ():
			_asyncons = [(each, 0)]
			break
			
		if each.handler is not None:			
			_asyncons.append ((each, each.handler.jobs ()))
	
	if _asyncons:
		_asyncons.sort (key = lambda x: x [1])		
		if _asyncons [0][1] < http2.MAX_HTTP2_CONCURRENT_STREAMS:
			asyncon = _asyncons [0][0]
	
	return asyncon
	

class SocketPool:
	object_timeout = 120
	maintern_interval = 30
	use_syn_connection = False
	ctype = None

	def __init__ (self, logger, backend = False, use_pool = True):
		self.__socketfarm = {}
		self.__protos = {}
		self.__numget  = 0
		self.__last_maintern = time.time ()
		self.logger = logger
		self.backend = backend
		self.use_pool = use_pool
		self.lock = threading.RLock ()
		self.numobj = 0
	
	def match (self, request):
		return False		
	
	def get_name (self):
		return "__socketpool__"
				
	def status (self):
		info = {}
		cluster = {}
		info ["numget"] = self.__numget
				
		try:
			with self.lock:
				for serverkey, node in list(self.__socketfarm.items ()):	
					nnode = {}
					nnode ["numactives"] = len ([x for x in list(node.values ()) if x.isactive ()])
					nnode ["numconnecteds"] = len ([x for x in list(node.values ()) if x.isconnected ()])
					conns = []
					for asyncon in list(node.values ()):
						stu = {
							"class": asyncon.__class__.__name__, 
							"connected": asyncon.isconnected (), 
							"isactive": asyncon.isactive (), 
							"request_count": asyncon.get_request_count (),
							"event_time": time.asctime (time.localtime (asyncon.event_time)), 
							"zombie_timeout": asyncon.zombie_timeout,								
						}
						try: stu ["has_result"] = asyncon.has_result
						except AttributeError: pass						
						
						if hasattr (asyncon, "get_history"):
							stu ["history"] = asyncon.get_history ()
							
						try:
							stu ["in_map"] = asyncon.is_channel_in_map ()
						except AttributeError: 
							pass	
							
						conns.append (stu)
													
					nnode ["connections"] = conns
					cluster [serverkey] = nnode
			
		except:
			self.logger.trace ()
					
		info ["cluster"] = cluster
		return info
		
	def report (self, asyncon, well_functioning):
		pass # for competitable
	
	def get_nodes (self):
		with self.lock:
			if not self.__socketfarm: return [None] # at least one item needs
			return list(self.__socketfarm.items ())
		
	def maintern (self):
		# close unused sockets				
		for serverkey, node in list (self.__socketfarm.items ()):
			for _id, asyncon in list (node.items ()):
				if not hasattr (asyncon, "maintern"):
					continue
				
				try:
					deletable = asyncon.maintern (self.object_timeout)
				except:
					self.logger.trace ()
				else:
					if deletable:
						asyncon.handler = None # break back ref.
						try: del node [_id]
						except KeyError: pass
						self.numobj -= 1
			
			# KeyError, WHY?
			if not node:
				try: del self.__socketfarm [serverkey]
				except KeyError: pass						
				try: del self.__protos [serverkey]
				except KeyError: pass
		
		self.logger ('mainterned %d socket pool' % len (self.__socketfarm), 'info')
				
	def _get (self, serverkey, server, *args):
		asyncon = None	
		if self.use_pool and time.time () - self.__last_maintern > self.maintern_interval:
			self.__last_maintern = time.time ()
			try:				
				self.maintern ()
			except:
				self.logger.trace ()
					
		self.__numget += 1
		if not self.use_pool:
			asyncon = self.create_asyncon (server, *args)
			
		else:	
			if serverkey not in self.__socketfarm:
				self.__socketfarm [serverkey] = {}
				asyncon = self.create_asyncon (server, *args)
				self.__socketfarm [serverkey][id (asyncon)] = asyncon
				
			else:		
				asyncons = list(self.__socketfarm [serverkey].values ())
				if self.__protos.get (serverkey) in PROTO_CONCURRENT_STREAMS:
					asyncon = select_channel (asyncons)
					
				else:
					for each in asyncons:								
						if not each.isactive ():
							asyncon = each
							break
				
				if not asyncon:
					asyncon = self.create_asyncon (server, *args)
					self.__socketfarm [serverkey][id (asyncon)] = asyncon
			
			proto = self.__protos.get (serverkey)
			if not proto and asyncon.get_proto ():
				self.__protos [serverkey] = asyncon.get_proto ()
				
		asyncon.set_active (True)		
		return asyncon
	
	def create_asyncon (self, server, scheme):		
		if scheme in ("https", "wss"):
			__conn_class = asynconnect.AsynSSLConnect
			__dft_Port = 443
		elif scheme == "tunnel":
			__conn_class = asynconnect.AsynConnect
			__dft_Port = 443
		elif scheme == "proxy":
			__conn_class = asynconnect.AsynConnect			
			__dft_Port = 5000
		elif scheme == "proxys":
			__conn_class = asynconnect.AsynSSLProxyConnect
			__dft_Port = 5000				
		else:
			__conn_class = asynconnect.AsynConnect
			__dft_Port = 80

		if self.use_syn_connection:
			if scheme == "https":
				__conn_class = synconnect.SynSSLConnect
			elif scheme == "http":	
				__conn_class = synconnect.SynConnect
		try:
			addr, port = server.split (":", 1)
			port = int (port)
		except ValueError:
			addr, port = server, __dft_Port
		
		self.numobj += 1			
		asyncon = __conn_class ((addr, port), self.lock, self.logger)	
		self.backend and asyncon.set_backend ()
		asyncon.set_proxy (scheme == "proxy")
		return asyncon
				
	def get (self, uri):		
		scheme, server, script, params, qs, fragment = urlparse (uri)
		serverkey = "%s://%s" % (scheme, server)
		try:
			with self.lock:
				return self._get (serverkey, server, scheme)
		except:
			self.logger.trace ()
			return None
			
	def noop (self):
		if not self.use_pool:
			return 
			
		with self.lock:
			for server in list(self.__socketfarm.keys ()):					
				for asyncon in list(self.__socketfarm [server].values ()):
					asyncon.set_event_time ()
			
	def cleanup (self):
		if not self.use_pool:
			return 
			
		try:
			with self.lock:
				for server in list(self.__socketfarm.keys ()):					
					for asyncon in list(self.__socketfarm [server].values ()):
						asyncon.disconnect ()
						del asyncon
				self.__socketfarm = {}
				self.__protos = {}
							
		except:
			self.logger.trace ()


pool = None

def create (logger, backend = False, use_pool = True):
	global pool
	if pool is None:
		pool = SocketPool (logger, backend, use_pool)

def get (uri):	
	return pool.get (uri)
		
def cleanup ():	
	pool.cleanup ()

def noop ():
	pool.noop ()
	


# -*- coding: utf-8 -*-
"""
hyper/http11/parser
~~~~~~~~~~~~~~~~~~~

This module contains hyper's pure-Python HTTP/1.1 parser. This module defines
an abstraction layer for HTTP/1.1 parsing that allows for dropping in other
modules if needed, in order to obtain speedups on your chosen platform.
"""
from collections import namedtuple


Response = namedtuple(
    'Response', ['status', 'msg', 'minor_version', 'headers', 'consumed']
)


class ParseError(Exception):
    """
    An invalid HTTP message was passed to the parser.
    """
    pass


class Parser(object):
    """
    A single HTTP parser object.
    This object is not thread-safe, and it does maintain state that is shared
    across parsing requests. For this reason, make sure that access to this
    object is synchronized if you use it across multiple threads.
    """
    def __init__(self):
        pass

    def parse_response(self, buffer):
        """
        Parses a single HTTP response from a buffer.
        :param buffer: A ``memoryview`` object wrapping a buffer containing a
            HTTP response.
        :returns: A :class:`Response <hyper.http11.parser.Response>` object, or
            ``None`` if there is not enough data in the buffer.
        """
        # Begin by copying the data out of the buffer. This is necessary
        # because as much as possible we want to use the built-in bytestring
        # methods, rather than looping over the data in Python.
        temp_buffer = buffer.tobytes()

        index = temp_buffer.find(b'\n')
        if index == -1:
            return None

        version, status, reason = temp_buffer[0:index].split(None, 2)
        if not version.startswith(b'HTTP/1.'):
            raise ParseError("Not HTTP/1.X!")

        minor_version = int(version[7:])
        status = int(status)
        reason = memoryview(reason.strip())

        # Chomp the newline.
        index += 1

        # Now, parse the headers out.
        end_index = index
        headers = []

        while True:
            end_index = temp_buffer.find(b'\n', index)
            if end_index == -1:
                return None
            elif (end_index - index) <= 1:
                # Chomp the newline
                end_index += 1
                break

            name, value = temp_buffer[index:end_index].split(b':', 1)
            value = value.strip()
            headers.append((memoryview(name), memoryview(value)))
            index = end_index + 1

        resp = Response(status, reason, minor_version, headers, end_index)
        return resp



from rs4 import asynchat, asyncore
import re, os, sys
import ssl
import socket
import time
import zlib
from warnings import warn
from errno import ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED, EWOULDBLOCK
if os.name == "nt":
	from errno import WSAENOTCONN
import select
import threading
from . import adns
from ..protocols.http2 import H2_PROTOCOLS
from ..athreads.fifo import await_fifo
from rs4.ssl_ import resolve_cert_reqs, resolve_ssl_version, create_urllib3_context
from collections import deque
from ..protocols.http import respcodes
import random

DEBUG = 0
DEFAULT_ZOMBIE_TIMEOUT = 60
DEFAULT_KEEP_ALIVE = 2

class SocketPanic (Exception): pass
class TimeOut (Exception): pass

class AsynConnect (asynchat.async_chat):
	ac_in_buffer_size = 65535
	ac_out_buffer_size = 65535
	zombie_timeout = DEFAULT_ZOMBIE_TIMEOUT
	keep_alive = DEFAULT_KEEP_ALIVE	
	request_count = 0
	active = 0
	fifo_class = deque
	keep_connect = True
	is_ssl = False
	
	def __init__ (self, address, lock = None, logger = None):
		self.address = address
		self.lock = lock
		self.logger = logger
		self._cv = threading.Condition ()
		self.__sendlock = None
		self.__no_more_request = False
		self.set_event_time ()
		self.handler = None
		
		self.auth = None
		self.proxy = False
		self.initialize_connection ()
		self._closed = False
		self.backend = False
		
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo = self.fifo_class ()
		asyncore.dispatcher.__init__(self)
	
	def __repr__ (self):
		return "<AsynConnect %s:%d>" % self.address
		
	def duplicate (self):
		new_asyncon = self.__class__ (self.address, self.lock, self.logger)
		# used in socketpool
		new_asyncon.proxy = self.proxy
		# used in skitai cluster manager
		new_asyncon.keep_alive = self.keep_alive
		new_asyncon.auth = self.auth
		new_asyncon.backend = self.backend
		new_asyncon.keep_connect = self.keep_connect
		return new_asyncon
	
	def set_backend (self, backend_keep_alive = 10):
		self.backend = True
		self.keep_alive = backend_keep_alive
		
	def set_auth (self, auth):
		self.auth = auth
	
	def get_auth (self):
		return self.auth
				
	def close (self):
		# sometimes socket is not closed at once
		# possibly related with SSL socket
		# then prevent doble callbacking in request_handler
				
		if self.socket:
			# self.socket is still None, when DNS not found
			asynchat.async_chat.close (self)
			self._fileno = None
			
		# re-init asychat
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo.clear()
		self._proto = None
		self._closed = True		
			
		if not self.handler:
			# return to the pool
			return self.set_active (False)
		
		if not self.errcode:
			# disconnect intentionally
			return
		
		handler, self.handller = self.handler, None
		keep_active = False
		try:
			keep_active = handler.connection_closed (self.errcode, self.errmsg)			
		except: 
			self.trace ()
		
		if not keep_active:
			#print (handler.request.meta ['sid'], 'not keepalive...')
			self.set_active (False)	
			if self.errcode not in (704, 712, 722):
				# controlled shutdown
				self.logger (
					".....socket %s has been closed (reason: %d)" % ("%s:%d" % self.address, self.errcode),
					"info"
				)
		# DO NOT Change any props, because may be request has been restarted	
		
	def end_tran (self):
		if DEBUG:
			self.logger ('end_tran {rid:%s} %s' % (self.handler.request.meta.get ('req_id', -1), self.handler.request.uri), 'debug')
		if not self.backend:			
			self.del_channel ()
		self.set_active (False)
		if not self.keep_connect:
			self.disconnect ()
				
	def use_sendlock (self):
		self.__sendlock = threading.Lock ()
		self.initiate_send = self._initiate_send_ts
		
	def _initiate_send_ts (self):
		with self.__sendlock:
			return asynchat.async_chat.initiate_send (self)
				
	def get_proto (self):
		with self.lock:
			p = self._proto
		return p
	
	def set_proto (self, proto):
		with self.lock:
			self._proto = proto		
				
	def get_history (self):
		return self.__history
				
	def initialize_connection (self):		
		self._raised_ENOTCONN = 0 # for win32
		self.__history = []
		self._proto = None
		self._handshaking = False
		self._handshaked = False
		
		self.established = False		
		self.upgraded = False		
		
	def set_event_time (self):
		self.event_time = time.time ()
	
	def is_proxy (self):
		return self.proxy
			
	def log (self, msg, logtype):
		if self.handler is not None and hasattr (self.handler, "log"):
			self.handler.log (msg, logtype)
		elif self.logger:
			self.logger (msg, logtype)
		else:
			warn ("No logger")
			
	def trace (self):
		if self.handler is not None and hasattr (self.handler, "trace"):
			self.handler.trace ()
		elif self.logger:
			self.logger.trace ()
		else:
			warn ("No logger for traceback")
		
	def clean_shutdown_control (self, phase, time_in_this_phase):	
		self.__no_more_request = True
		if self.isactive () or (self.handler and self.handler.working ()):		
			return 1
		else:				
			self.handle_close (712, "Controlled Shutdown")
			self.__no_more_request = False
			return 0
	
	def is_channel_in_map (self, map = None):
		if map is None:
			map = self._map		
		return self._fileno in map
		
	def set_active (self, flag, nolock = False):
		if flag:
			flag = time.time ()
		else:
			flag = 0
			
		if nolock or self.lock is None:
			self.active = flag
			if not flag: self.set_timeout (self.keep_alive)
			return
			
		self.lock.acquire ()
		self.active = flag
		self.request_count += 1		
		if not flag: self.set_timeout (self.keep_alive)
		self.lock.release ()
	
	def get_active (self, nolock = False):
		if nolock or self.lock is None:
			return self.active
		self.lock.acquire ()
		active = self.active
		self.lock.release ()	
		return active
	
	def isactive (self):	
		return self.get_active () > 0
	
	def isconnected (self):
		with self.lock:
			r = self.connected
		return r	
		
	def get_request_count (self):	
		return self.request_count
		
	def del_channel (self, map = None):
		fd = self._fileno
		if map is None:
			map = self._map
		if fd in map:
			del map[fd]
            			
	def create_socket (self, family, type):
		self.family_and_type = family, type
		sock = socket.socket (family, type)
		sock.setblocking (0)
		self.set_socket (sock)
			
	def connect (self):
		if adns.query:
			adns.query (self.address [0], "A", callback = self.continue_connect)
		else:			
			self.continue_connect ()
		
	def continue_connect (self, answer = None):
		self.initialize_connection ()
		if not adns.query:
			self.create_socket (socket.AF_INET, socket.SOCK_STREAM)		
			try: asynchat.async_chat.connect (self, self.address)
			except:	self.handle_error (714)
			return
		
		ipaddr = answer [-1]["data"]
		#print (self.handler.request.meta ['sid'], ipaddr, 'continue_connect...')
		if not ipaddr:			
			return self.handle_close (704)			
		else:	
			port = self.address [1]
		
		self.create_socket (socket.AF_INET, socket.SOCK_STREAM)		
		try: 
			asynchat.async_chat.connect (self, (ipaddr, port))
		except:	
			self.handle_error (714)
		
	def recv (self, buffer_size):
		try:
			data = self.socket.recv (buffer_size)
			if not data:
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''
			else:
				self.set_event_time ()
				return data		
		except socket.error as why:
			if why.errno in asyncore._DISCONNECTED:
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''				
			else:
				raise
	
	def send (self, data):
		try:
			numsent = self.socket.send (data)
			if numsent:
				self.set_event_time ()
			return numsent	
			
		except socket.error as why:
			if why.errno == EWOULDBLOCK:
				return 0				
			elif why.errno in asyncore._DISCONNECTED:
				#print (">>>>>>>>>> why.errno == asyncore.ENOTCONN", why.errno == asyncore.ENOTCONN)
				if os.name == "nt" and why.errno == asyncore.ENOTCONN:
					# winsock sometimes raise ENOTCONN and sometimes recovered.
					# Found this error at http://file.hungryboarder.com:8080/HBAdManager/pa.html?siteId=hboarder&zoneId=S-2
					if self._raised_ENOTCONN <= 3:
						self._raised_ENOTCONN += 1
						return 0
					else:
						self._raised_ENOTCONN = 0
				self.handle_close (700, "Connection closed unexpectedly in send")
				return 0
			else:
				raise
	
	def close_if_over_keep_live (self):		
		if time.time () - self.event_time > self.keep_alive:
			self.disconnect ()
		
	def set_timeout (self, timeout = 10):
		# CAUTION: used at proxy.tunnel_handler
		self.zombie_timeout = timeout		
	
	def set_keep_alive (self, keep_alive = 10):
		self.keep_alive = keep_alive
		
	def handle_connect (self):		
		if hasattr (self.handler, "has_been_connected"):		
			self.handler.has_been_connected ()
					
	def handle_expt (self):
		self.handle_close (703)
	
	def handle_error (self, code = 701):
		self.trace ()
		self.handle_close (code)
	
	def handle_timeout (self):		
		self.handle_close (702)
		
	def handle_expt_event (self):		
		err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
		if err != 0:
			self.handle_close (703, "Socket %d Error" % err)
		else:
			self.handle_expt ()
	
	def maintern (self, object_timeout):		
		if time.time () - self.event_time > object_timeout:
			if self.handler:
				if hasattr (self.handler, "control_shutdown"):
					self.handler.control_shutdown ()
				self.handle_close (722)
			else:	
				self.disconnect ()
			return True
		return False
		
	# proxy POST need no init_send
	def push (self, thing, init_send = True):
		if self.connected:
			self.close_if_over_keep_live () # check keep-alive
			
		if isinstance(thing, (bytes, bytearray, memoryview)):
			asynchat.async_chat.push (self, thing)
		else:
			self.push_with_producer (thing, init_send)	
		
	def push_with_producer (self, producer, init_send = True):
		if self.connected:
			self.close_if_over_keep_live () # check keep-alive
			
		self.producer_fifo.append (producer)
		if init_send:
			self.initiate_send ()
	
	def handle_abort (self):
		self.handler = None
		self.close ()		
		
	def handle_close (self, code = 700, msg = ""):
		if code == 0: msg = ""
		self.errcode = code
		if msg:
			self.errmsg = msg
		else:
			self.errmsg = respcodes.get (code, "Undefined Error")			
		self.close ()
							
	def collect_incoming_data (self, data):
		if not self.handler:
			if self.connected:
				self.logger ("recv data but no handler, droping data %d" % len (data), "warn")
				self.disconnect ()
			return
		self.handler.collect_incoming_data (data)
	
	def found_terminator (self):
		if not self.handler:
			if self.connected:
				self.logger ("found terminator but no handler", "warn")
				self.disconnect ()
			return # already closed
		self.handler.found_terminator ()
	
	def disconnect (self):
		# no error
		self.handle_close (0)
	
	def reconnect (self):
		self.disconnect ()		
		self.connect ()
	
	def set_proxy (self, flag = True):
		self.proxy = flag
	
	def begin_tran (self, handler):
		if self.__no_more_request:
			return self.handle_close (705)
		
		# IMP: the reason why DNS error, _closed must be located here
		self._closed = False
		self.errcode = 0
		self.errmsg = ""
		
		self.handler = handler
		if DEBUG:
			self.logger ('begin_tran {rid:%s} %s' % (self.handler.request.meta.get ('req_id', -1), self.handler.request.uri), 'debug')
		
		self.set_event_time ()
		# IMP: call add_channel () AFTER push()	otherwise threading issue will be raised
		try:
			if not self.connected:				
				#print (self.handler.request.meta ['sid'], 'connecting...')
				self.connect ()
								
			elif not self.backend:
				#print (self.handler.request.meta ['sid'], 'add channel...')
				self.add_channel ()
				
		except:
			self.handle_error ()


class AsynSSLConnect (AsynConnect):	
	is_ssl = True

	def __init__ (self, address, lock = None, logger = None):
		AsynConnect.__init__ (self, address, lock, logger)
		self.ac_negotiate_http2 = True
	
	def negotiate_http2 (self, flag):
		self.ac_negotiate_http2 = flag
		
	def handshake (self):		
		if not self._handshaking:
			err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
			if err != 0:
				raise OSError(err, asyncore._strerror(err))
				
			ssl_context = create_urllib3_context(ssl_version=resolve_ssl_version(None), cert_reqs=resolve_cert_reqs(None))
			if self.ac_negotiate_http2:								
				try: ssl_context.set_alpn_protocols (H2_PROTOCOLS)
				except AttributeError: ssl_context.set_npn_protocols (H2_PROTOCOLS)								
			self.socket = ssl_context.wrap_socket (self.socket, do_handshake_on_connect = False, server_hostname = self.address [0])			
			self._handshaking = True
			
		try:
			self.socket.do_handshake ()
		except ssl.SSLError as why:
			if why.args [0] in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE):
				return False
			raise ssl.SSLError(why)
			
		try: self._proto = self.socket.selected_alpn_protocol()
		except (AttributeError, NotImplementedError): 
			try: self._proto = self.socket.selected_npn_protocol()
			except (AttributeError, NotImplementedError): pass

		self._handshaked = True		
		return True
							
	def handle_connect_event (self):
		try:
			if not self._handshaked and not self.handshake ():
				return
		except:
			return self.handle_error (713)		
		AsynConnect.handle_connect_event (self)		
		
	def recv (self, buffer_size):
		if self._closed:
			# usually handshaking failure, already handled exception
			return b''
			
		try:
			data = self.socket.recv (buffer_size)			
			if not data:				
				self.handle_close (700, "Connection closed unexpectedly")
				return b''
			else:				
				self.set_event_time ()
				return data
			
		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_READ:
				try: 
					raise BlockingIOError				
				except NameError:
					raise socket.error (EWOULDBLOCK)

			# closed connection
			elif why.errno in (ssl.SSL_ERROR_ZERO_RETURN, ssl.SSL_ERROR_EOF):
				self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN or SSL_ERROR_EOF")
				return b''
				
			else:
				raise

	def send (self, data):		
		if self._closed:
			# usually handshaking failure, already handled exception
			return
			
		try:
			numsent = self.socket.send (data)	
			if numsent:
				self.set_event_time ()
			return numsent	

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_WRITE:
				return 0
			elif why.errno == ssl.SSL_ERROR_ZERO_RETURN:				
				self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN")
				return 0
			else:
				raise


class AsynSSLProxyConnect (AsynSSLConnect, AsynConnect):
	is_ssl = True

	def __init__ (self, address, lock = None, logger = None):
		AsynConnect.__init__ (self, address, lock, logger)

	def handle_connect_event (self):
		if self.established:
			AsynSSLConnect.handle_connect_event (self)
		else:	
			AsynConnect.handle_connect_event (self)
	
	def recv (self, buffer_size):		
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.recv (self, buffer_size)				
		else:
			return AsynConnect.recv (self, buffer_size)
			
	def send (self, data):	
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.send (self, data)
		else:
			return AsynConnect.send (self, data)

	


# -*- coding: utf-8 -*-
"""
h2/errors
~~~~~~~~~~~~~~~~~~~

Global error code registry containing the established HTTP/2 error codes.

The current registry is available at:
https://tools.ietf.org/html/rfc7540#section-11.4
"""
import enum


class ErrorCodes(enum.IntEnum):
    """
    All known HTTP/2 error codes.

    .. versionadded:: 2.5.0
    """
    #: Graceful shutdown.
    NO_ERROR = 0x0

    #: Protocol error detected.
    PROTOCOL_ERROR = 0x1

    #: Implementation fault.
    INTERNAL_ERROR = 0x2

    #: Flow-control limits exceeded.
    FLOW_CONTROL_ERROR = 0x3

    #: Settings not acknowledged.
    SETTINGS_TIMEOUT = 0x4

    #: Frame received for closed stream.
    STREAM_CLOSED = 0x5

    #: Frame size incorrect.
    FRAME_SIZE_ERROR = 0x6

    #: Stream not processed.
    REFUSED_STREAM = 0x7

    #: Stream cancelled.
    CANCEL = 0x8

    #: Compression state not updated.
    COMPRESSION_ERROR = 0x9

    #: TCP connection error for CONNECT method.
    CONNECT_ERROR = 0xa

    #: Processing capacity exceeded.
    ENHANCE_YOUR_CALM = 0xb

    #: Negotiated TLS parameters not acceptable.
    INADEQUATE_SECURITY = 0xc

    #: Use HTTP/1.1 for the request.
    HTTP_1_1_REQUIRED = 0xd


def _error_code_from_int(code):
    """
    Given an integer error code, returns either one of :class:`ErrorCodes
    <h2.errors.ErrorCodes>` or, if not present in the known set of codes,
    returns the integer directly.
    """
    try:
        return ErrorCodes(code)
    except ValueError:
        return code


#: Graceful shutdown.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.NO_ERROR
#:    <h2.errors.ErrorCodes.NO_ERROR>`.
NO_ERROR = ErrorCodes.NO_ERROR

#: Protocol error detected.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.PROTOCOL_ERROR
#:    <h2.errors.ErrorCodes.PROTOCOL_ERROR>`.
PROTOCOL_ERROR = ErrorCodes.PROTOCOL_ERROR

#: Implementation fault.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.INTERNAL_ERROR
#:    <h2.errors.ErrorCodes.INTERNAL_ERROR>`.
INTERNAL_ERROR = ErrorCodes.INTERNAL_ERROR

#: Flow-control limits exceeded.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.FLOW_CONTROL_ERROR
#:    <h2.errors.ErrorCodes.FLOW_CONTROL_ERROR>`.
FLOW_CONTROL_ERROR = ErrorCodes.FLOW_CONTROL_ERROR

#: Settings not acknowledged.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.SETTINGS_TIMEOUT
#:    <h2.errors.ErrorCodes.SETTINGS_TIMEOUT>`.
SETTINGS_TIMEOUT = ErrorCodes.SETTINGS_TIMEOUT

#: Frame received for closed stream.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.STREAM_CLOSED
#:    <h2.errors.ErrorCodes.STREAM_CLOSED>`.
STREAM_CLOSED = ErrorCodes.STREAM_CLOSED

#: Frame size incorrect.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.FRAME_SIZE_ERROR
#:    <h2.errors.ErrorCodes.FRAME_SIZE_ERROR>`.
FRAME_SIZE_ERROR = ErrorCodes.FRAME_SIZE_ERROR

#: Stream not processed.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.REFUSED_STREAM
#:    <h2.errors.ErrorCodes.REFUSED_STREAM>`.
REFUSED_STREAM = ErrorCodes.REFUSED_STREAM

#: Stream cancelled.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.CANCEL
#:    <h2.errors.ErrorCodes.CANCEL>`.
CANCEL = ErrorCodes.CANCEL

#: Compression state not updated.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.COMPRESSION_ERROR
#:    <h2.errors.ErrorCodes.COMPRESSION_ERROR>`.
COMPRESSION_ERROR = ErrorCodes.COMPRESSION_ERROR

#: TCP connection error for CONNECT method.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.CONNECT_ERROR
#:    <h2.errors.ErrorCodes.CONNECT_ERROR>`.
CONNECT_ERROR = ErrorCodes.CONNECT_ERROR

#: Processing capacity exceeded.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.ENHANCE_YOUR_CALM
#:    <h2.errors.ErrorCodes.ENHANCE_YOUR_CALM>`.
ENHANCE_YOUR_CALM = ErrorCodes.ENHANCE_YOUR_CALM

#: Negotiated TLS parameters not acceptable.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.INADEQUATE_SECURITY
#:    <h2.errors.ErrorCodes.INADEQUATE_SECURITY>`.
INADEQUATE_SECURITY = ErrorCodes.INADEQUATE_SECURITY

#: Use HTTP/1.1 for the request.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes.HTTP_1_1_REQUIRED
#:    <h2.errors.ErrorCodes.HTTP_1_1_REQUIRED>`.
HTTP_1_1_REQUIRED = ErrorCodes.HTTP_1_1_REQUIRED

#: All known HTTP/2 error codes.
#:
#: .. deprecated:: 2.5.0
#:    Deprecated in favour of :class:`ErrorCodes <h2.errors.ErrorCodes>`.
H2_ERRORS = list(ErrorCodes)

__all__ = ['H2_ERRORS', 'NO_ERROR', 'PROTOCOL_ERROR', 'INTERNAL_ERROR',
           'FLOW_CONTROL_ERROR', 'SETTINGS_TIMEOUT', 'STREAM_CLOSED',
           'FRAME_SIZE_ERROR', 'REFUSED_STREAM', 'CANCEL', 'COMPRESSION_ERROR',
           'CONNECT_ERROR', 'ENHANCE_YOUR_CALM', 'INADEQUATE_SECURITY',
           'HTTP_1_1_REQUIRED', 'ErrorCodes']



# -*- coding: utf-8 -*-
"""
hyper/cli
~~~~~~~~~

Command line interface for Hyper inspired by Httpie.
"""
import json
import locale
import logging
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from argparse import OPTIONAL, ZERO_OR_MORE
from pprint import pformat
from textwrap import dedent

from hyper import HTTPConnection, HTTP20Connection
from hyper import __version__
from hyper.compat import is_py2, urlencode, urlsplit, write_to_stdout
from hyper.common.util import to_host_port_tuple


log = logging.getLogger('hyper')

PREFERRED_ENCODING = locale.getpreferredencoding()

# Various separators used in args
SEP_HEADERS = ':'
SEP_QUERY = '=='
SEP_DATA = '='

SEP_GROUP_ITEMS = [
    SEP_HEADERS,
    SEP_QUERY,
    SEP_DATA,
]


class KeyValue(object):
    """Base key-value pair parsed from CLI."""

    def __init__(self, key, value, sep, orig):
        self.key = key
        self.value = value
        self.sep = sep
        self.orig = orig


class KeyValueArgType(object):
    """A key-value pair argument type used with `argparse`.

    Parses a key-value arg and constructs a `KeyValue` instance.
    Used for headers, form data, and other key-value pair types.
    This class is inspired by httpie and implements simple tokenizer only.
    """
    def __init__(self, *separators):
        self.separators = separators

    def __call__(self, string):
        for sep in self.separators:
            splitted = string.split(sep, 1)
            if len(splitted) == 2:
                key, value = splitted
                return KeyValue(key, value, sep, string)


def make_positional_argument(parser):
    parser.add_argument(
        'method', metavar='METHOD', nargs=OPTIONAL, default='GET',
        help=dedent("""
        The HTTP method to be used for the request
        (GET, POST, PUT, DELETE, ...).
        """))
    parser.add_argument(
        '_url', metavar='URL',
        help=dedent("""
        The scheme defaults to 'https://' if the URL does not include one.
        """))
    parser.add_argument(
        'items',
        metavar='REQUEST_ITEM',
        nargs=ZERO_OR_MORE,
        type=KeyValueArgType(*SEP_GROUP_ITEMS),
        help=dedent("""
        Optional key-value pairs to be included in the request.
        The separator used determines the type:

        ':' HTTP headers:

            Referer:http://httpie.org  Cookie:foo=bar  User-Agent:bacon/1.0

        '==' URL parameters to be appended to the request URI:

            search==hyper

        '=' Data fields to be serialized into a JSON object:

            name=Hyper  language=Python  description='CLI HTTP client'
        """))


def make_troubleshooting_argument(parser):
    parser.add_argument(
        '--version', action='version', version=__version__,
        help='Show version and exit.')
    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='Show debugging information (loglevel=DEBUG)')
    parser.add_argument(
        '--h2', action='store_true', default=False,
        help="Do HTTP/2 directly, skipping plaintext upgrade and ignoring "
             "NPN/ALPN."
    )


def split_host_and_port(hostname):
    if ':' in hostname:
        return to_host_port_tuple(hostname, default_port=443)
    return hostname, None


class UrlInfo(object):
    def __init__(self):
        self.fragment = None
        self.host = 'localhost'
        self.netloc = None
        self.path = '/'
        self.port = 443
        self.query = None
        self.scheme = 'https'
        self.secure = False


def set_url_info(args):
    info = UrlInfo()
    _result = urlsplit(args._url)
    for attr in vars(info).keys():
        value = getattr(_result, attr, None)
        if value:
            setattr(info, attr, value)

    if info.scheme == 'http' and not _result.port:
        info.port = 80

    # Set the secure arg is the scheme is HTTPS, otherwise do unsecured.
    info.secure = info.scheme == 'https'

    if info.netloc:
        hostname, _ = split_host_and_port(info.netloc)
        info.host = hostname  # ensure stripping port number
    else:
        if _result.path:
            _path = _result.path.split('/', 1)
            hostname, port = split_host_and_port(_path[0])
            info.host = hostname
            if info.path == _path[0]:
                info.path = '/'
            elif len(_path) == 2 and _path[1]:
                info.path = '/' + _path[1]
            if port is not None:
                info.port = port

    log.debug('Url Info: %s', vars(info))
    args.url = info


def set_request_data(args):
    body, headers, params = {}, {}, {}
    for i in args.items:
        if i.sep == SEP_HEADERS:
            if i.key:
                headers[i.key] = i.value
            else:
                # when overriding a HTTP/2 special header there will be a
                # leading colon, which tricks the command line parser into
                # thinking the header is empty
                k, v = i.value.split(':', 1)
                headers[':' + k] = v
        elif i.sep == SEP_QUERY:
            params[i.key] = i.value
        elif i.sep == SEP_DATA:
            value = i.value
            if is_py2:  # pragma: no cover
                value = value.decode(PREFERRED_ENCODING)
            body[i.key] = value

    if params:
        args.url.path += '?' + urlencode(params)

    if body:
        content_type = 'application/json'
        headers.setdefault('content-type', content_type)
        args.body = json.dumps(body)

    if args.method is None:
        args.method = 'POST' if args.body else 'GET'

    args.method = args.method.upper()
    args.headers = headers


def parse_argument(argv=None):
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter)
    parser.set_defaults(body=None, headers={})
    make_positional_argument(parser)
    make_troubleshooting_argument(parser)
    args = parser.parse_args(sys.argv[1:] if argv is None else argv)

    if args.debug:
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        log.addHandler(handler)
        log.setLevel(logging.DEBUG)

    set_url_info(args)
    set_request_data(args)
    return args


def get_content_type_and_charset(response):
    charset = 'utf-8'
    content_type = response.headers.get('content-type')
    if content_type is None:
        return 'unknown', charset

    content_type = content_type[0].decode('utf-8').lower()
    type_and_charset = content_type.split(';', 1)
    ctype = type_and_charset[0].strip()
    if len(type_and_charset) == 2:
        charset = type_and_charset[1].strip().split('=')[1]

    return ctype, charset


def request(args):
    if not args.h2:
        conn = HTTPConnection(
            args.url.host, args.url.port, secure=args.url.secure
        )
    else:  # pragma: no cover
        conn = HTTP20Connection(
            args.url.host,
            args.url.port,
            secure=args.url.secure,
            force_proto='h2'
        )

    conn.request(args.method, args.url.path, args.body, args.headers)
    response = conn.get_response()
    log.debug('Response Headers:\n%s', pformat(response.headers))
    ctype, charset = get_content_type_and_charset(response)
    data = response.read()
    return data


def main(argv=None):
    args = parse_argument(argv)
    log.debug('Commandline Argument: %s', args)
    data = request(args)
    write_to_stdout(data)


if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2014 Rackspace
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from .compat import to_bytes
from .misc import NON_PCT_ENCODED


def normalize_scheme(scheme):
    return scheme.lower()


def normalize_authority(authority):
    userinfo, host, port = authority
    result = ''
    if userinfo:
        result += normalize_percent_characters(userinfo) + '@'
    if host:
        result += host.lower()
    if port:
        result += ':' + port
    return result


def normalize_path(path):
    if not path:
        return path

    path = normalize_percent_characters(path)
    return remove_dot_segments(path)


def normalize_query(query):
    return normalize_percent_characters(query)


def normalize_fragment(fragment):
    return normalize_percent_characters(fragment)


PERCENT_MATCHER = re.compile('%[A-Fa-f0-9]{2}')


def normalize_percent_characters(s):
    """All percent characters should be upper-cased.

    For example, ``"%3afoo%DF%ab"`` should be turned into ``"%3Afoo%DF%AB"``.
    """
    matches = set(PERCENT_MATCHER.findall(s))
    for m in matches:
        if not m.isupper():
            s = s.replace(m, m.upper())
    return s


def remove_dot_segments(s):
    # See http://tools.ietf.org/html/rfc3986#section-5.2.4 for pseudo-code
    segments = s.split('/')  # Turn the path into a list of segments
    output = []  # Initialize the variable to use to store output

    for segment in segments:
        # '.' is the current directory, so ignore it, it is superfluous
        if segment == '.':
            continue
        # Anything other than '..', should be appended to the output
        elif segment != '..':
            output.append(segment)
        # In this case segment == '..', if we can, we should pop the last
        # element
        elif output:
            output.pop()

    # If the path starts with '/' and the output is empty or the first string
    # is non-empty
    if s.startswith('/') and (not output or output[0]):
        output.insert(0, '')

    # If the path starts with '/.' or '/..' ensure we add one more empty
    # string to add a trailing '/'
    if s.endswith(('/.', '/..')):
        output.append('')

    return '/'.join(output)


def encode_component(uri_component, encoding):
    if uri_component is None:
        return uri_component

    uri_bytes = to_bytes(uri_component, encoding)

    encoded_uri = bytearray()

    for i in range(0, len(uri_bytes)):
        # Will return a single character bytestring on both Python 2 & 3
        byte = uri_bytes[i:i+1]
        byte_ord = ord(byte)
        if byte_ord < 128 and byte.decode() in NON_PCT_ENCODED:
            encoded_uri.extend(byte)
            continue
        encoded_uri.extend('%{0:02x}'.format(byte_ord).encode())

    return encoded_uri.decode(encoding)



# -*- coding: utf-8 -*-
"""
hyper/http20/window
~~~~~~~~~~~~~~~~~~~

Objects that understand flow control in hyper.

HTTP/2 implements connection- and stream-level flow control. This flow
control is mandatory. Unfortunately, it's difficult for hyper to be
all that intelligent about how it manages flow control in a general case.

This module defines an interface for pluggable flow-control managers. These
managers will define a flow-control policy. This policy will determine when to
send WINDOWUPDATE frames.
"""


class BaseFlowControlManager(object):
    """
    The abstract base class for flow control managers.

    This class defines the interface for pluggable flow-control managers. A
    flow-control manager defines a flow-control policy, which basically boils
    down to deciding when to increase the flow control window.

    This decision can be based on a number of factors:

    - the initial window size,
    - the size of the document being retrieved,
    - the size of the received data frames,
    - any other information the manager can obtain

    A flow-control manager may be defined at the connection level or at the
    stream level. If no stream-level flow-control manager is defined, an
    instance of the connection-level flow control manager is used.

    A class that inherits from this one must not adjust the member variables
    defined in this class. They are updated and set by methods on this class.
    """
    def __init__(self, initial_window_size, document_size=None):
        #: The initial size of the connection window in bytes. This is set at
        #: creation time.
        self.initial_window_size = initial_window_size

        #: The current size of the connection window. Any methods overridden
        #: by the user must not adjust this value.
        self.window_size = initial_window_size

        #: The size of the document being retrieved, in bytes. This is
        #: retrieved from the Content-Length header, if provided. Note that
        #: the total number of bytes that will be received may be larger than
        #: this value due to HTTP/2 padding. It should not be assumed that
        #: simply because the the document size is smaller than the initial
        #: window size that there will never be a need to increase the window
        #: size.
        self.document_size = document_size

    def increase_window_size(self, frame_size):
        """
        Determine whether or not to emit a WINDOWUPDATE frame.

        This method should be overridden to determine, based on the state of
        the system and the size of the received frame, whether or not a
        WindowUpdate frame should be sent for the stream.

        This method should *not* adjust any of the member variables of this
        class.

        Note that this method is called before the window size is decremented
        as a result of the frame being handled.

        :param frame_size: The size of the received frame. Note that this *may*
          be zero. When this parameter is zero, it's possible that a
          WINDOWUPDATE frame may want to be emitted anyway. A zero-length frame
          size is usually associated with a change in the size of the receive
          window due to a SETTINGS frame.
        :returns: The amount to increase the receive window by. Return zero if
          the window should not be increased.
        """
        raise NotImplementedError(
            "FlowControlManager is an abstract base class"
        )

    def blocked(self):
        """
        Called whenever the remote endpoint reports that it is blocked behind
        the flow control window.

        When this method is called the remote endpoint is signaling that it
        has more data to send and that the transport layer is capable of
        transmitting it, but that the HTTP/2 flow control window prevents it
        being sent.

        This method should return the size by which the window should be
        incremented, which may be zero. This method should *not* adjust any
        of the member variables of this class.

        :returns: The amount to increase the receive window by. Return zero if
          the window should not be increased.
        """
        # TODO: Is this method necessary?
        raise NotImplementedError(
            "FlowControlManager is an abstract base class"
        )

    def _handle_frame(self, frame_size):
        """
        This internal method is called by the connection or stream that owns
        the flow control manager. It handles the generic behaviour of flow
        control managers: namely, keeping track of the window size.
        """
        rc = self.increase_window_size(frame_size)
        self.window_size -= frame_size
        self.window_size += rc
        return rc

    def _blocked(self):
        """
        This internal method is called by the connection or stream that owns
        the flow control manager. It handles the generic behaviour of receiving
        BLOCKED frames.
        """
        rc = self.blocked()
        self.window_size += rc
        return rc


class FlowControlManager(BaseFlowControlManager):
    """
    ``hyper``'s default flow control manager.

    This implements hyper's flow control algorithms. This algorithm attempts to
    reduce the number of WINDOWUPDATE frames we send without blocking the
    remote endpoint behind the flow control window.

    This algorithm will become more complicated over time. In the current form,
    the algorithm is very simple:

    - When the flow control window gets less than 1/4 of the maximum size,
      increment back to the maximum.
    - Otherwise, if the flow control window gets to less than 1kB, increment
      back to the maximum.
    """
    def increase_window_size(self, frame_size):
        future_window_size = self.window_size - frame_size

        if ((future_window_size < (self.initial_window_size / 4)) or
                (future_window_size < 1000)):
            return self.initial_window_size - future_window_size

        return 0

    def blocked(self):
        return self.initial_window_size - self.window_size



# -*- coding: utf-8 -*-
"""
hyper/common/connection
~~~~~~~~~~~~~~~~~~~~~~~

Hyper's HTTP/1.1 and HTTP/2 abstraction layer.
"""
from .exceptions import TLSUpgrade, HTTPUpgrade
from ..http11.connection import HTTP11Connection
from ..http20.connection import HTTP20Connection
from ..tls import H2_NPN_PROTOCOLS, H2C_PROTOCOL


class HTTPConnection(object):
    """
    An object representing a single HTTP connection to a server.

    This object behaves similarly to the Python standard library's
    ``HTTPConnection`` object, with a few critical differences.

    Most of the standard library's arguments to the constructor are not
    supported by hyper. Most optional parameters apply to *either* HTTP/1.1 or
    HTTP/2.

    :param host: The host to connect to. This may be an IP address or a
        hostname, and optionally may include a port: for example,
        ``'http2bin.org'``, ``'http2bin.org:443'`` or ``'127.0.0.1'``.
    :param port: (optional) The port to connect to. If not provided and one
        also isn't provided in the ``host`` parameter, defaults to 80.
    :param secure: (optional) Whether the request should use TLS.
        Defaults to ``False`` for most requests, but to ``True`` for any
        request issued to port 443.
    :param window_manager: (optional) The class to use to manage flow control
        windows. This needs to be a subclass of the
        :class:`BaseFlowControlManager
        <hyper.http20.window.BaseFlowControlManager>`. If not provided,
        :class:`FlowControlManager <hyper.http20.window.FlowControlManager>`
        will be used.
    :param enable_push: (optional) Whether the server is allowed to push
        resources to the client (see
        :meth:`get_pushes() <hyper.HTTP20Connection.get_pushes>`).
    :param ssl_context: (optional) A class with custom certificate settings.
        If not provided then hyper's default ``SSLContext`` is used instead.
    :param proxy_host: (optional) The proxy to connect to.  This can be an IP
        address or a host name and may include a port.
    :param proxy_port: (optional) The proxy port to connect to. If not provided
        and one also isn't provided in the ``proxy`` parameter, defaults to
        8080.
    """
    def __init__(self,
                 host,
                 port=None,
                 secure=None,
                 window_manager=None,
                 enable_push=False,
                 ssl_context=None,
                 proxy_host=None,
                 proxy_port=None,
                 **kwargs):

        self._host = host
        self._port = port
        self._h1_kwargs = {
            'secure': secure, 'ssl_context': ssl_context,
            'proxy_host': proxy_host, 'proxy_port': proxy_port
        }
        self._h2_kwargs = {
            'window_manager': window_manager, 'enable_push': enable_push,
            'secure': secure, 'ssl_context': ssl_context,
            'proxy_host': proxy_host, 'proxy_port': proxy_port
        }

        # Add any unexpected kwargs to both dictionaries.
        self._h1_kwargs.update(kwargs)
        self._h2_kwargs.update(kwargs)

        self._conn = HTTP11Connection(
            self._host, self._port, **self._h1_kwargs
        )

    def request(self, method, url, body=None, headers=None):
        """
        This will send a request to the server using the HTTP request method
        ``method`` and the selector ``url``. If the ``body`` argument is
        present, it should be string or bytes object of data to send after the
        headers are finished. Strings are encoded as UTF-8. To use other
        encodings, pass a bytes object. The Content-Length header is set to the
        length of the body field.

        :param method: The request method, e.g. ``'GET'``.
        :param url: The URL to contact, e.g. ``'/path/segment'``.
        :param body: (optional) The request body to send. Must be a bytestring
            or a file-like object.
        :param headers: (optional) The headers to send on the request.
        :returns: A stream ID for the request, or ``None`` if the request is
            made over HTTP/1.1.
        """

        headers = headers or {}

        try:
            return self._conn.request(
                method=method, url=url, body=body, headers=headers
            )
        except TLSUpgrade as e:
            # We upgraded in the NPN/ALPN handshake. We can just go straight to
            # the world of HTTP/2. Replace the backing object and insert the
            # socket into it.
            assert e.negotiated in H2_NPN_PROTOCOLS

            self._conn = HTTP20Connection(
                self._host, self._port, **self._h2_kwargs
            )
            self._conn._sock = e.sock

            # Because we skipped the connecting logic, we need to send the
            # HTTP/2 preamble.
            self._conn._send_preamble()

            return self._conn.request(
                method=method, url=url, body=body, headers=headers
            )

    def get_response(self, *args, **kwargs):
        """
        Returns a response object.
        """
        try:
            return self._conn.get_response(*args, **kwargs)
        except HTTPUpgrade as e:
            # We upgraded via the HTTP Upgrade mechanism. We can just
            # go straight to the world of HTTP/2. Replace the backing object
            # and insert the socket into it.
            assert e.negotiated == H2C_PROTOCOL

            self._conn = HTTP20Connection(
                self._host, self._port, **self._h2_kwargs
            )

            self._conn._connect_upgrade(e.sock)
            # stream id 1 is used by the upgrade request and response
            # and is half-closed by the client

            return self._conn.get_response(1)

    # The following two methods are the implementation of the context manager
    # protocol.
    def __enter__(self):  # pragma: no cover
        return self

    def __exit__(self, type, value, tb):  # pragma: no cover
        self._conn.close()
        return False  # Never swallow exceptions.

    # Can anyone say 'proxy object pattern'?
    def __getattr__(self, name):
        return getattr(self._conn, name)
